package me.hammerle.supersnuvi.gamelogic;

import java.io.File;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;
import java.util.stream.Collectors;
import javafx.scene.paint.Color;
import me.hammerle.supersnuvi.entity.Entity;
import me.hammerle.supersnuvi.entity.Hero;
import me.hammerle.supersnuvi.input.IKeyHandler;
import me.hammerle.supersnuvi.rendering.IGameRenderer;
import me.hammerle.supersnuvi.tiles.BottledSoulTile;
import me.hammerle.supersnuvi.tiles.Location;
import me.hammerle.supersnuvi.tiles.StartTile;
import me.hammerle.supersnuvi.tiles.Tile;
import me.hammerle.supersnuvi.util.CollisionBox;

public final class Level 
{
    public final static double MAX_ENERGY = 100.0;
    
    private final StateRenderer state;
    private final IGameRenderer renderer;
    private final Color TRANSPARENT_WHITE = new Color(1, 1, 1, 0.5);
    
    private final boolean worldLoaded;
    private final LevelData data;
    
    private final String name;
    
    private final HashMap<Integer, Entity> entities;
    private Hero hero;
    private int entityCounter;
    
    private boolean shouldReset;
    private boolean done;
    
    private int souls;
    private int maxSouls;
    
    private TreeSet<Point> spawns;
    
    private double energy;
    
    public Level(StateRenderer state, File f)
    {
        this.state = state;
        this.renderer = state.getRenderer();
        this.data = new LevelData(f);
        this.worldLoaded = data.load();
        this.entities = new HashMap<>();
        this.entityCounter = 0;
        this.shouldReset = false;
        this.done = false;
        
        this.spawns = new TreeSet<>();
        
        if(worldLoaded)
        {
            this.name = data.getString("name", "error");
            maxSouls = 0;
            data.forEachInteractTile((x, y, tile) -> 
            {
                if(tile == BottledSoulTile.ID)
                {
                    maxSouls++;
                }
                else if(tile == StartTile.ID)
                {
                    spawns.add(new Point(x, y));
                }
            });
        }
        else
        {
            this.name = "error";
            maxSouls = 0;
        }
        
        // there must be at least one spawn
        if(spawns.isEmpty())
        {
            spawns.add(new Point(5, 5));
        }
        
        resetLevel();
    }
    
    // -------------------------------------------------------------------------
    // basic stuff
    // -------------------------------------------------------------------------

    public String getName() 
    {
        return name;
    }
    
    public void finishLevel()
    {
        shouldReset = true;
        done = true;
    }
    
    public boolean shouldFinish()
    {
        return done;
    }
    
    public boolean shouldReset()
    {
        return shouldReset;
    }
    
    public void scheduleReset()
    {
        shouldReset = true;
    }
    
    public void resetLevel()
    {
        state.resetTiles();
        resetEnergy();
        souls = 0;
        shouldReset = false;
        done = false;
        Hero h = spawnHero();
        hero = h;
        entities.clear();
        entities.put(entityCounter++, h);
    }
    
    public Hero spawnHero()
    {
        Hero h = (Hero) entities.values().stream().filter(ent -> ent instanceof Hero).findFirst().orElse(null);
        if(h == null || hero.getX() < 0)
        {
            // first spawn or out of map, use first spawn
            Point p = spawns.first();
            return new Hero(this, renderer.toCoord(p.getX()), renderer.toCoord(p.getY()));
        }
        else
        {
            // hero is somewhere in the map, getting last spawn
            Point p = spawns.floor(new Point(renderer.toBlock(hero.getX()), renderer.toBlock(hero.getY())));
            if(p == null)
            {
                p = spawns.first();
            }
            return new Hero(this, renderer.toCoord(p.getX()), renderer.toCoord(p.getY()));
        }
    }
    
    public void increaseSouls()
    {
        souls++;
    }
    
    public void resetEnergy()
    {
        energy = MAX_ENERGY;
    }
    
    public void addEnergy(double d)
    {
        energy += d;
        if(energy > MAX_ENERGY)
        {
            energy = MAX_ENERGY;
        }
        else if(energy < 0.0)
        {
            energy = 0.0;
        }
    }
    
    public double getEnergy()
    {
        return energy;
    }
    
    // -------------------------------------------------------------------------
    // tick
    // -------------------------------------------------------------------------
    
    public void tick(IKeyHandler keys)
    {
        if(worldLoaded)
        {
            // setting the right view center
            double rWidth = renderer.getWidth();
            double rHeight = renderer.getHeight();
            
            double centerX = Math.min(Math.max(0, hero.getX() - rWidth / 2), data.getWidth() * 32 - rWidth);
            double centerY = Math.min(Math.max(0, hero.getY() - rHeight / 2), data.getHeight() * 32 - rHeight);
            
            renderer.setViewCenter(centerX, centerY);
            
            // end
            
            if(hero.getY() < 0)
            {
                resetLevel();
                return;
            }
            
            state.tick();
            
            // doing entity logic first
            entities.values().forEach(entity -> 
            {
                entity.controlTick(keys);
                entity.resetMovementData();
                entity.tick();
            });

            int startX = renderer.getFirstVisibleBlockX();
            int startY = renderer.getFirstVisibleBlockY();
            int endX = Math.min(renderer.getLastVisibleBlockX() + 1, data.getWidth());
            int endY = Math.min(renderer.getLastVisibleBlockY() + 1, data.getHeight());
            
            data.forEachBackground((x, y, tile) ->
            {
                if(tile != -1)
                {
                    Tile t = state.getTile(tile);
                    renderer.drawBlockImage(t.getImage(x, y), x, y, t.getRenderOffsetX(), t.getRenderOffsetY());
                }
            }, startX, endX, startY, endY);
            
            // render entities here
            entities.values().forEach(en -> 
            {
                if(en.drawImageFlipped())
                {
                    renderer.drawFlippedImage(en.getImage(), en.getRenderX() + en.getRenderOffsetX(), en.getRenderY() + en.getRenderOffsetY());
                }
                else
                {
                    renderer.drawImage(en.getImage(), en.getRenderX() + en.getRenderOffsetX(), en.getRenderY() + en.getRenderOffsetY());
                }
            });
            // end entity rendering
            
            data.forEachForeground((x, y, tile) ->
            {
                if(tile != -1)
                {
                    Tile t = state.getTile(tile);
                    renderer.drawBlockImage(t.getImage(x, y), x, y, t.getRenderOffsetX(), t.getRenderOffsetY());
                }
            }, startX, endX, startY, endY);
            
            // rendering souls
            String s = souls + "/" + maxSouls;
            double w = renderer.getStringWidth(s) + BottledSoulTile.IMAGE.getWidth();
            renderer.fillRec(0, 0, w, BottledSoulTile.IMAGE.getHeight(), TRANSPARENT_WHITE);
            
            renderer.drawFixedImage(BottledSoulTile.IMAGE, 0, 0);
            renderer.prepareStringDrawing(Color.BLACK);            
            double d = (BottledSoulTile.IMAGE.getHeight() - renderer.getLineHeight()) / 2;
            renderer.drawString(s, BottledSoulTile.IMAGE.getWidth(), d);
            renderer.stopStringDrawing();
            
            // energy
            double energyWidth = renderer.getWidth() * 0.2;
            double ex = (renderer.getWidth() / 2) - energyWidth;
            energyWidth *= 2;
            renderer.fillRec(ex, 5, energyWidth, 20, Color.BLACK);
            energyWidth -= 4;
            renderer.fillRec(ex + 2, 7, energyWidth * (energy / MAX_ENERGY), 16, Color.BLUE);
            
            if(!hero.isMoving())
            {
                addEnergy(0.1);
            }
        }
    }   
    
    // -------------------------------------------------------------------------
    // collision box, interaction layer
    // -------------------------------------------------------------------------
    
    public Tile getInteractionTile(int x, int y)
    {
        int i = data.getInteractionTile(x, y);
        if(i == -1)
        {
            return StateRenderer.FALLBACK_TILE;
        }
        return state.getInteractionTile(i);
    }
    
    public CollisionBox getTileMovementBox(int x, int y)
    {
        int i = data.getInteractionTile(x, y);
        if(i == -1)
        {
            return CollisionBox.NULL_BOX;
        }
        Tile tile = state.getInteractionTile(i);
        return tile.getMovementBox(x, y).reset().offset(renderer.toCoord(x), renderer.toCoord(y));
    }
    
    public LinkedList<Location> getTilesInMovementOf(CollisionBox cb)
    {
        LinkedList<Location> boxes = new LinkedList<>();
        int startX = renderer.toBlock(cb.getMinX());
        int endX = renderer.toBlock(cb.getMaxX());
        int startY = renderer.toBlock(cb.getMinY());
        int endY = renderer.toBlock(cb.getMaxY());
        
        for(int x = startX; x <= endX; x++)
        {
            for(int y = startY; y <= endY; y++)
            {
                if(getTileMovementBox(x, y).intersects(cb))
                {
                    boxes.add(new Location(getInteractionTile(x, y), x, y));
                }
            }
        }
        return boxes;
    }
    
    public CollisionBox getTileCollisionBox(int x, int y)
    {
        int i = data.getInteractionTile(x, y);
        if(i == -1)
        {
            return CollisionBox.NULL_BOX;
        }
        Tile tile = state.getInteractionTile(i);
        return tile.getCollisionBox(x, y).reset().offset(renderer.toCoord(x), renderer.toCoord(y));
    }
    
    public LinkedList<Location> getTilesCollidingWith(CollisionBox cb)
    {
        LinkedList<Location> boxes = new LinkedList<>();
        int startX = renderer.toBlock(cb.getMinX());
        int endX = renderer.toBlock(cb.getMaxX());
        int startY = renderer.toBlock(cb.getMinY());
        int endY = renderer.toBlock(cb.getMaxY());
        
        for(int x = startX; x <= endX; x++)
        {
            for(int y = startY; y <= endY; y++)
            {
                if(getTileCollisionBox(x, y).intersects(cb))
                {
                    boxes.add(new Location(getInteractionTile(x, y), x, y));
                }
            }
        }
        return boxes;
    }
    
    public List<Entity> getEntitiesCollidingWith(Entity not, CollisionBox cb)
    {
        return entities.values().stream().filter(ent -> ent != not && ent.getBox().intersects(cb)).collect(Collectors.toList());
    }
    
    public LinkedList<CollisionBox> getMovementBoxesAt(Entity not, CollisionBox cb)
    {
        LinkedList<CollisionBox> boxes = new LinkedList<>(entities.values().stream()
                        .filter(ent -> ent != not && ent.getBox().intersects(cb))
                        .map(ent -> ent.getBox())
                        .collect(Collectors.toList()));
        int startX = renderer.toBlock(cb.getMinX());
        int endX = renderer.toBlock(cb.getMaxX());
        int startY = renderer.toBlock(cb.getMinY());
        int endY = renderer.toBlock(cb.getMaxY());
        
        CollisionBox box;
        for(int x = startX; x <= endX; x++)
        {
            for(int y = startY; y <= endY; y++)
            {
                box = getTileMovementBox(x, y);
                if(box.intersects(cb))
                {
                    boxes.add(box.copy());
                }
            }
        }
        return boxes;
    }
}
